/* bzflag
 * Copyright (c) 1993 - 2003 Tim Riker
 *
 * This package is free software;  you can redistribute it and/or
 * modify it under the terms of the license found in the file
 * named LICENSE that should have accompanied this file.
 *
 * THIS PACKAGE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

/* 3dTYpes:
 *	classes for vertex vector and matrix stuff
 */

#ifndef _3DTYPES_H_
#define _3DTYPES_H_

#ifndef UINT
#define UINT	unsigned int
#endif

#ifndef DEG_RAD
#define DEG_RAD 0.017453f
#endif

	// 3d variables
	typedef struct
	{
		float	x,y,z;
	}Vertex3D;

	typedef struct
	{
		float	s,t;
	}Vertex2D;

	typedef struct
	{
		float	r,g,b;
	}Color;

	typedef struct
	{
		long			aVerts[3];
		long			aNorms[3];
		long			aTCs[3];
		Vertex3D		rFaceNormal;
	}Triangle;

	#include "math.h"

	/*--------------------------------- Vertex3D operators -----------------------------------------*/

	/*--------------------------------- operator - -----------------------------------------*/

	inline Vertex3D operator - (const Vertex3D &p0)
	{
		Vertex3D p1;

		p1.x = - p0.x;
		p1.y = - p0.y;
		p1.z = - p0.z;

		return p1;
	}

	/*--------------------------------- operator + -----------------------------------------*/

	inline Vertex3D operator + (const Vertex3D &p0, const Vertex3D &p1)
	{
		Vertex3D p2;

		p2.x = p0.x + p1.x;
		p2.y = p0.y + p1.y;
		p2.z = p0.z + p1.z;

		return p2;
	}

	/*--------------------------------- operator - -----------------------------------------*/

	inline Vertex3D operator - (const Vertex3D &p0, const Vertex3D &p1)
	{
		Vertex3D p2;

		p2.x = p0.x - p1.x;
		p2.y = p0.y - p1.y;
		p2.z = p0.z - p1.z;

		return p2;
	}

	/*--------------------------------- operator * -----------------------------------------*/

	inline Vertex3D operator * (const Vertex3D &p0, float scalar)
	{
		Vertex3D p2;

		p2.x = p0.x * scalar;
		p2.y = p0.y * scalar;
		p2.z = p0.z * scalar;

		return p2;
	}

	/*--------------------------------- operator / -----------------------------------------*/

	inline Vertex3D operator / (const Vertex3D &p0, float scalar)
	{
		Vertex3D p2;

		scalar = 1.0f / scalar;
		p2.x = p0.x * scalar;
		p2.y = p0.y * scalar;
		p2.z = p0.z * scalar;

		return p2;
	}

	/*--------------------------------- operator * -----------------------------------------*/

	inline Vertex3D operator * (float scalar, const Vertex3D &p0)
	{
		Vertex3D p2;

		p2.x = p0.x * scalar;
		p2.y = p0.y * scalar;
		p2.z = p0.z * scalar;

		return p2;
	}

	/*--------------------------------- operator / -----------------------------------------*/

	inline Vertex3D operator / (float scalar, const Vertex3D &p0)
	{
		Vertex3D p2;

		scalar = 1.0f / scalar;
		p2.x = p0.x * scalar;
		p2.y = p0.y * scalar;
		p2.z = p0.z * scalar;

		return p2;
	}

	/*--------------------------------- operator += -----------------------------------------*/

	inline void operator += (Vertex3D &p0, const Vertex3D &p1)
	{
		p0.x += p1.x;
		p0.y += p1.y;
		p0.z += p1.z;
	}

	/*--------------------------------- operator -= -----------------------------------------*/

	inline void operator -= (Vertex3D &p0, const Vertex3D &p1)
	{
		p0.x -= p1.x;
		p0.y -= p1.y;
		p0.z -= p1.z;
	}

	/*--------------------------------- operator *= -----------------------------------------*/

	inline void operator *= (Vertex3D &p0, const Vertex3D &p1)
	{
		p0.x *= p1.x;
		p0.y *= p1.y;
		p0.z *= p1.z;
	}

	/*--------------------------------- operator /= -----------------------------------------*/

	inline void operator /= (Vertex3D &p0, const Vertex3D &p1)
	{
		p0.x /= p1.x;
		p0.y /= p1.y;
		p0.z /= p1.z;
	}

	/*--------------------------------- operator *= -----------------------------------------*/

	inline void operator *= (Vertex3D &p0, float scalar)
	{
		p0.x *= scalar;
		p0.y *= scalar;
		p0.z *= scalar;
	}

	/*--------------------------------- operator /= -----------------------------------------*/

	inline void operator /= (Vertex3D &p0, float scalar)
	{
		scalar = 1.0f / scalar;
		p0.x *= scalar;
		p0.y *= scalar;
		p0.z *= scalar;
	}

	/*------------------------------------- dot ---------------------------------------------*/

	inline float dot(const Vertex3D &lhs, const Vertex3D &rhs)
	{
		return lhs.x * rhs.x + lhs.y * rhs.y + lhs.z * rhs.z;
	}

	/*------------------------------------- dot_2d ---------------------------------------------*/

	inline float dot_2d(const Vertex3D &lhs, const Vertex3D &rhs)
	{
		return lhs.x * rhs.x + lhs.y * rhs.y;
	}

	/*------------------------------------- cross ---------------------------------------------*/

	inline Vertex3D& cross(const Vertex3D &lhs, const Vertex3D &rhs, Vertex3D &result)
	{
		float temp_x, temp_y;		// this allows result to be same as lhs or rhs

		temp_x = lhs.z * rhs.y - lhs.y * rhs.z;
		temp_y = lhs.z * rhs.x - lhs.x * rhs.z;
		result.z = lhs.x * rhs.y - lhs.y * rhs.x;
		result.x = temp_x;
		result.y = temp_y;
		return result;
	}

	/*------------------------------------- cross_2d ---------------------------------------------*/

	inline float cross_2d(const Vertex3D &lhs, const Vertex3D &rhs)
	{
		return lhs.x * rhs.y - lhs.y * rhs.x;
	}

	/*------------------------------------- get_mag ---------------------------------------------*/

	inline float get_mag(const Vertex3D &vec)
	{
		return (float)sqrt(dot(vec, vec));
	}

	/*------------------------------------- get_mag_2d ---------------------------------------------*/

	inline float get_mag_2d(const Vertex3D &vec)
	{
		return (float)sqrt(dot_2d(vec, vec));
	}

	/*------------------------------------- set_mag ---------------------------------------------*/

	inline bool set_mag(Vertex3D &vec, float mag)
	{
		float old_mag = get_mag(vec);
		if (old_mag < 1.0e-15)
		{
			vec.x = mag;
			vec.z = vec.y = 0;
			return false;
		}
		vec *= mag / old_mag;
		return true;
	}

	/*------------------------------------- set_mag_2d ---------------------------------------------*/

	inline bool set_mag_2d(Vertex3D &vec, float mag)
	{
		float old_mag = get_mag_2d(vec);
		if (old_mag < 1.0e-15)
		{
			vec.x = mag;
			vec.z = vec.y = 0;
			return false;
		}
		mag /= old_mag;
		vec.x *= mag;
		vec.y *= mag;
		return true;
	}
	
	#define XFORM_TRANS		1
	#define XFORM_SCALE		2
	#define XFORM_ROT		4
#ifdef WIN32
	#pragma pack (push, matrix_def, 1)			// Matrix is 1 byte aligned
#endif
	typedef struct
	{
		short flags;
		float el[3][4];
	}
	Matrix;
#ifdef WIN32
	#pragma pack (pop, matrix_def)				// restore struct alignment
#endif
	typedef enum
	{
		eXY, 
		eXZ, 
		eYZ,
		eXR,
		eXD,
		eNonPrimary
	}
	tePlanes;

	Matrix *ClassifyMatrix(Matrix*);
	Matrix *NormalizeMatrix(Matrix*);
	Matrix *GetIdentityMatrix(Matrix*);
	bool MatrixPlanar(const Matrix*, bool*);
	tePlanes MatrixPlane(const Matrix*);
	bool MatrixScaleNonUniform(const Matrix*);
	bool MatrixScaleNonUniform2d(const Matrix*);
	Vertex3D *MatrixScale(const Matrix*, Vertex3D*);
	bool EqualMatrix(const Matrix*, const Matrix*);
	bool SamePlane(const Matrix*, const Matrix*, bool*);
	Matrix *TranslateMatrix(Matrix*, const Vertex3D*);
	Matrix *TranslationMatrix(Matrix*, const Vertex3D*);
	Matrix *ScaleMatrix(Matrix*, const Vertex3D*);
	Matrix *ScalingMatrix(Matrix*, const Vertex3D*);
	Matrix *RotationMatrix(Matrix*, const Vertex3D*, float, float);
	Matrix *RotationMatrix(Matrix*, float, float);
	Matrix *Rotate2dMatrix(Matrix*, float, const Vertex3D*);
	Matrix *MirrorMatrix(Matrix*, const Vertex3D*, const Vertex3D*);
	Matrix *ForceDepthMatrix(Matrix*, double);
	Matrix *ConcatMatrices(Matrix*, const Matrix*);
	Matrix *InvertMatrix(Matrix*);
	Vertex3D *TransformPos(Vertex3D*, const Matrix*);
	Vertex3D *TransformPos2d(Vertex3D*, const Matrix*);
	Vertex3D *TransformNorm(Vertex3D*, const Matrix*);
	Vertex3D *TransformNorm2d(Vertex3D*, const Matrix*);
	Matrix *MatrixFromAxes(Matrix*, const Vertex3D*, const Vertex3D*, const Vertex3D*);
	Matrix *DifferenceMatrix(Matrix*, const Vertex3D*, const Vertex3D*, const Vertex3D*, const Vertex3D*, const Vertex3D*, const Vertex3D*);

#endif//_3DTYPES_H_
